# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2020 Philipp Wolfer
# Copyright (C) 2020-2021 Laurent Monin
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from PyQt5.QtGui import QColor

from test.picardtestcase import PicardTestCase

from picard import config

from picard.ui.colors import (
    InterfaceColors,
    UnknownColorException,
    interface_colors,
)


settings = {
    'interface_colors': {
        'unknowncolor': '#deadbe',
        'entity_error': '#abcdef',
    },
    'interface_colors_dark': {
        'unknowncolor': '#deadbe',
        'entity_error': '#abcdef',
    }
}


class InterfaceColorsTest(PicardTestCase):
    def setUp(self):
        super().setUp()
        self.set_config_values(settings)

    def test_interface_colors(self):
        for key in ('interface_colors', 'interface_colors_dark'):
            interface_colors = InterfaceColors(dark_theme=key == 'interface_colors_dark')
            with self.assertRaises(UnknownColorException):
                interface_colors.get_color('testcolor')
            default_colors = interface_colors.default_colors
            self.assertEqual(interface_colors.get_color('entity_error'), default_colors['entity_error'].value)
            interface_colors.load_from_config()
            self.assertEqual(interface_colors.get_color('entity_error'), '#abcdef')
            self.assertEqual(interface_colors.get_colors()['entity_error'], '#abcdef')
            interface_colors.set_color('entity_error', '#000000')
            self.assertTrue(interface_colors.save_to_config())
            self.assertEqual(config.setting[key]['entity_error'], '#000000')
            self.assertNotIn('unknowncolor', config.setting[key])
            self.assertEqual(interface_colors.get_color_description('entity_error'), default_colors['entity_error'].description)
            self.assertEqual(interface_colors.get_qcolor('entity_error'), QColor('#000000'))

    def test_interface_colors_default(self):
        self.assertIsInstance(interface_colors, InterfaceColors)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2021 Bob Swift
# Copyright (C) 2021 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


import datetime

import yaml

from test.picardtestcase import PicardTestCase

from picard.script.serializer import (
    FileNamingScript,
    PicardScript,
    ScriptImportError,
)


class _DateTime(datetime.datetime):
    @classmethod
    def utcnow(cls):
        return cls(year=2020, month=1, day=2, hour=12, minute=34, second=56, microsecond=789, tzinfo=None)


class PicardScriptTest(PicardTestCase):

    original_datetime = datetime.datetime

    def setUp(self):
        super().setUp()
        # Save original datetime object and substitute one returning
        # a fixed utcnow() value for testing.
        datetime.datetime = _DateTime

    def tearDown(self):
        # Restore original datetime object
        datetime.datetime = self.original_datetime

    def assertYamlEquals(self, yaml_str, obj, msg=None):
        self.assertEqual(obj, yaml.safe_load(yaml_str), msg)

    def test_script_object_1(self):
        # Check initial loaded values.
        test_script = PicardScript(title='Script 1', script='Script text', id='12345', last_updated='2021-04-26', script_language_version='1.0')
        self.assertEqual(test_script.id, '12345')
        self.assertEqual(test_script['id'], '12345')
        self.assertEqual(test_script.last_updated, '2021-04-26')
        self.assertEqual(test_script['last_updated'], '2021-04-26')
        self.assertYamlEquals(test_script.to_yaml(), {"id": "12345", "script": "Script text\n", "script_language_version": "1.0", "title": "Script 1"})

    def test_script_object_2(self):
        # Check updating values directly so as not to modify `last_updated`.
        test_script = PicardScript(title='Script 1', script='Script text', id='12345', last_updated='2021-04-26')
        test_script.id = '54321'
        self.assertEqual(test_script.id, '54321')
        self.assertEqual(test_script['id'], '54321')
        self.assertEqual(test_script.last_updated, '2021-04-26')
        self.assertEqual(test_script['last_updated'], '2021-04-26')

        test_script.title = 'Updated Script 1'
        self.assertEqual(test_script.title, 'Updated Script 1')
        self.assertEqual(test_script['title'], 'Updated Script 1')
        self.assertEqual(test_script['last_updated'], '2021-04-26')

    def test_script_object_3(self):
        # Check updating values that are ignored from modifying `last_updated`.
        test_script = PicardScript(title='Script 1', script='Script text', id='12345', last_updated='2021-04-26')
        test_script.update_script_setting(id='54321')
        self.assertEqual(test_script.id, '54321')
        self.assertEqual(test_script['id'], '54321')
        self.assertEqual(test_script.last_updated, '2021-04-26')
        self.assertEqual(test_script['last_updated'], '2021-04-26')

    def test_script_object_4(self):
        # Check updating values that modify `last_updated`.
        test_script = PicardScript(title='Script 1', script='Script text', id='12345', last_updated='2021-04-26')
        test_script.update_script_setting(title='Updated Script 1')
        self.assertEqual(test_script.title, 'Updated Script 1')
        self.assertEqual(test_script['title'], 'Updated Script 1')
        self.assertEqual(test_script.last_updated, '2020-01-02 12:34:56 UTC')
        self.assertEqual(test_script['last_updated'], '2020-01-02 12:34:56 UTC')

    def test_script_object_5(self):
        # Check updating values from dict that modify `last_updated`.
        test_script = PicardScript(title='Script 1', script='Script text', id='12345', last_updated='2021-04-26')
        test_script.update_from_dict({"script": "Updated script"})
        self.assertEqual(test_script.script, 'Updated script')
        self.assertEqual(test_script['script'], 'Updated script')
        self.assertEqual(test_script.last_updated, '2020-01-02 12:34:56 UTC')
        self.assertEqual(test_script['last_updated'], '2020-01-02 12:34:56 UTC')

    def test_script_object_6(self):
        # Test that extra (unknown) settings are ignored during updating
        test_script = PicardScript(title='Script 1', script='Script text', id='12345', last_updated='2021-04-26', script_language_version='1.0')
        test_script.update_script_setting(description='Updated description')
        self.assertEqual(test_script['last_updated'], '2021-04-26')
        self.assertYamlEquals(test_script.to_yaml(), {"id": "12345", "script": "Script text\n", "script_language_version": "1.0", "title": "Script 1"})
        with self.assertRaises(AttributeError):
            print(test_script.description)

    def test_script_object_7(self):
        # Test that extra (unknown) settings are ignored during updating from dict
        test_script = PicardScript(title='Script 1', script='Script text', id='12345', last_updated='2021-04-26', script_language_version='1.0')
        test_script.update_from_dict({"description": "Updated description"})
        self.assertEqual(test_script['last_updated'], '2021-04-26')
        self.assertYamlEquals(test_script.to_yaml(), {"id": "12345", "script": "Script text\n", "script_language_version": "1.0", "title": "Script 1"})
        with self.assertRaises(AttributeError):
            print(test_script.description)

    def test_script_object_8(self):
        # Test that requested unknown settings return None
        test_script = PicardScript(title='Script 1', script='Script text', id='12345', last_updated='2021-04-26')
        self.assertEqual(test_script['unknown_setting'], None)

    def test_script_object_9(self):
        # Test that an exception is raised when creating or updating using an invalid YAML string
        with self.assertRaises(ScriptImportError):
            PicardScript().create_from_yaml('Not a YAML string')
        PicardScript(title='Script 1', script='Script text', id='12345', last_updated='2021-04-26', script_language_version='1.0')

    def test_naming_script_object_1(self):
        # Check initial loaded values.
        test_script = FileNamingScript(
            title='Script 1', script='Script text', id='12345', last_updated='2021-04-26',
            description='Script description', author='Script author', script_language_version='1.0'
        )
        self.assertEqual(test_script.id, '12345')
        self.assertEqual(test_script['id'], '12345')
        self.assertEqual(test_script.last_updated, '2021-04-26')
        self.assertEqual(test_script['last_updated'], '2021-04-26')
        self.assertEqual(test_script.script, 'Script text')
        self.assertEqual(test_script['script'], 'Script text')
        self.assertEqual(test_script.author, 'Script author')
        self.assertEqual(test_script['author'], 'Script author')
        self.assertEqual(
            test_script.to_yaml(),
            "title: Script 1\n"
            "description: |\n"
            "  Script description\n"
            "author: Script author\n"
            "license: ''\n"
            "version: ''\n"
            "last_updated: '2021-04-26'\n"
            "script_language_version: '1.0'\n"
            "script: |\n"
            "  Script text\n"
            "id: '12345'\n"
        )



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2023 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.

from collections import (
    defaultdict,
    deque,
    namedtuple,
)
from functools import partial
from typing import (
    Dict,
    List,
)

from PyQt5.QtNetwork import QNetworkReply

from picard.acoustid.json_helpers import (
    parse_recording,
    recording_has_metadata,
)
from picard.webservice import WebService
from picard.webservice.api_helpers import MBAPIHelper


# Only do extra lookup for recordings without metadata, if they have at least
# this percentage of sources compared to the recording with most sources.
SOURCE_THRESHOLD_NO_METADATA = 0.25

# Load max. this number of recordings without metadata per AcoustID
MAX_NO_METADATA_RECORDINGS = 3


class Recording:
    recording: dict
    result_score: float
    sources: int

    def __init__(self, recording, result_score=1.0, sources=1):
        self.recording = recording
        self.result_score = result_score
        self.sources = sources


IncompleteRecording = namedtuple('Recording', 'mbid acoustid result_score sources')


class RecordingResolver:
    """Given an AcoustID lookup result returns a list of MB recordings.
    The recordings are either directly taken from the AcoustID result or, if the
    results return only the MBID without metadata, loaded via the MB web service.
    """

    _recording_map: Dict[str, Dict[str, Recording]]

    def __init__(self, ws: WebService, doc: dict, callback: callable) -> None:
        self._mbapi = MBAPIHelper(ws)
        self._doc = doc
        self._callback = callback
        self._recording_map = defaultdict(dict)
        self._recording_cache = dict()
        self._missing_metadata = deque()

    def resolve(self) -> None:
        results = self._doc.get('results') or []
        incomplete_counts = defaultdict(lambda: 0)
        for result in results:
            recordings = result.get('recordings') or []
            result_score = get_score(result)
            acoustid = result.get('id')
            max_sources = max_source_count_raw_recording(recordings)
            for recording in sorted(recordings, key=lambda r: r.get('sources', 1), reverse=True):
                mbid = recording.get('id')
                sources = recording.get('sources', 1)
                if recording_has_metadata(recording):
                    mb_recording = parse_recording(recording)
                    self._recording_cache[mbid] = mb_recording
                    self._recording_map[acoustid][recording['id']] = Recording(
                        recording=mb_recording,
                        result_score=result_score,
                        sources=sources,
                    )
                else:
                    if (sources / max_sources > SOURCE_THRESHOLD_NO_METADATA
                        and incomplete_counts[acoustid] < MAX_NO_METADATA_RECORDINGS):
                        self._missing_metadata.append(IncompleteRecording(
                            mbid=mbid,
                            acoustid=acoustid,
                            result_score=result_score,
                            sources=sources,
                        ))
                        incomplete_counts[acoustid] += 1

        self._load_recordings()

    def _load_recordings(self):
        if not self._missing_metadata:
            self._send_results()
            return

        mbid = self._missing_metadata[0].mbid
        if mbid in self._recording_cache:
            mb_recording = self._recording_cache[mbid]
            self._recording_request_finished(mbid, mb_recording, None, None)
        else:
            self._mbapi.get_track_by_id(
                self._missing_metadata[0].mbid,
                partial(self._recording_request_finished, mbid),
                inc=('artists', 'release-groups', 'releases', 'media'),
            )

    def _recording_request_finished(self, original_mbid, mb_recording, http, error):
        recording = self._missing_metadata.popleft()
        if error:
            if error == QNetworkReply.NetworkError.ContentNotFoundError:
                # Recording does not exist, ignore and move on
                self._load_recordings()
            else:
                self._send_results(error)
            return

        mbid = mb_recording.get('id')
        recording_dict = self._recording_map[recording.acoustid]
        if mbid:
            self._recording_cache[mbid] = mb_recording
            # This was a redirect, cache the old MBID as well
            if original_mbid != mbid:
                self._recording_cache[original_mbid] = mb_recording
            if mbid not in recording_dict:
                recording_dict[mbid] = Recording(
                    recording=mb_recording,
                    result_score=recording.result_score,
                    sources=recording.sources,
                )
            else:
                recording_dict[mbid].sources += recording.sources
        self._load_recordings()

    def _send_results(self, error=None):
        self._callback(list(parse_recording_map(self._recording_map)), error)


def get_score(node):
    try:
        return float(node.get('score', 1.0))
    except (TypeError, ValueError):
        return 1.0


def parse_recording_map(recording_map: Dict[str, Dict[str, Recording]]):
    for acoustid, recordings in recording_map.items():
        recording_list = recordings.values()
        max_sources = max_source_count(recording_list)
        for recording in recording_list:
            parsed_recording = recording.recording
            if parsed_recording is not None:
                # Calculate a score based on result score and sources for this
                # recording relative to other recordings in this result
                score = min(recording.sources / max_sources, 1.0) * 100
                parsed_recording['score'] = score * recording.result_score
                parsed_recording['acoustid'] = acoustid
                parsed_recording['sources'] = recording.sources
            yield parsed_recording


def max_source_count(recordings: List[Recording]):
    """Given a list of recordings return the highest number of sources.
    This ignores recordings without metadata.
    """
    sources = {r.sources for r in recordings}
    sources.add(1)
    return max(sources)


def max_source_count_raw_recording(recordings: List[dict]):
    """Given a list of recordings return the highest number of sources.
    This ignores recordings without metadata.
    """
    sources = {r.get('sources', 1) for r in recordings}
    sources.add(1)
    return max(sources)



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2017 Sambhav Kothari
# Copyright (C) 2017, 2019-2022 Laurent Monin
# Copyright (C) 2018 Wieland Hoffmann
# Copyright (C) 2018-2023 Philipp Wolfer
# Copyright (C) 2020 dukeyin
# Copyright (C) 2021 Bob Swift
# Copyright (C) 2021 Vladislav Karbovskii
# Copyright (C) 2023 David Kellner
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from test.picardtestcase import (
    PicardTestCase,
    load_test_json,
)

from picard import config
from picard.album import Album
from picard.const import (
    ALIAS_TYPE_ARTIST_NAME_ID,
    ALIAS_TYPE_LEGAL_NAME_ID,
    ALIAS_TYPE_SEARCH_HINT_ID,
)
from picard.mbjson import (
    _locales_from_aliases,
    _node_skip_empty_iter,
    _parse_attributes,
    _relations_to_metadata_target_type_url,
    _translate_artist_node,
    artist_to_metadata,
    countries_from_node,
    get_score,
    label_info_from_node,
    media_formats_from_node,
    medium_to_metadata,
    recording_to_metadata,
    release_dates_and_countries_from_node,
    release_group_to_metadata,
    release_to_metadata,
    track_to_metadata,
)
from picard.metadata import Metadata
from picard.releasegroup import ReleaseGroup
from picard.track import Track


settings = {
    "standardize_tracks": False,
    "standardize_artists": False,
    "standardize_releases": False,
    "translate_artist_names": True,
    "translate_artist_names_script_exception": False,
    "standardize_instruments": True,
    "release_ars": True,
    "preferred_release_countries": [],
    "artist_locales": ['en'],
}


class MBJSONItersTest(PicardTestCase):
    def test_node_skip_empty_iter(self):
        d = {
            'bool_false': False,
            'bool_true': True,
            'int_0': 0,
            'int_1': 1,
            'float_0': 0.0,
            'float_1': 1.1,
            'list_empty': [],
            'list_non_empty': ['a'],
            'dict_empty': {},
            'dict_non_empty': {'a': 'b'},
        }
        expected = set(d) - {'list_empty', 'dict_empty'}
        result = set({k: v for k, v in _node_skip_empty_iter(d)})
        self.assertSetEqual(expected, result)


class MBJSONTest(PicardTestCase):
    def setUp(self):
        super().setUp()
        self.init_test(self.filename)

    def init_test(self, filename):
        self.set_config_values(settings)
        self.json_doc = load_test_json(filename)


class ReleaseTest(MBJSONTest):

    filename = 'release.json'

    def test_release(self):
        m = Metadata()
        a = Album("1")
        release_to_metadata(self.json_doc, m, a)
        self.assertEqual(m['album'], 'The Dark Side of the Moon')
        self.assertEqual(m['albumartist'], 'Pink Floyd')
        self.assertEqual(m['albumartistsort'], 'Pink Floyd')
        self.assertEqual(m['asin'], 'b123')
        self.assertEqual(m['barcode'], '123')
        self.assertEqual(m['catalognumber'], 'SHVL 804')
        self.assertEqual(m['date'], '1973-03-24')
        self.assertEqual(m['label'], 'Harvest')
        self.assertEqual(m['musicbrainz_albumartistid'], '83d91898-7763-47d7-b03b-b92132375c47')
        self.assertEqual(m['musicbrainz_albumid'], 'b84ee12a-09ef-421b-82de-0441a926375b')
        self.assertEqual(m['producer'], 'Hipgnosis')
        self.assertEqual(m['releasecountry'], 'GB')
        self.assertEqual(m['releasestatus'], 'official')
        self.assertEqual(m['script'], 'Latn')
        self.assertEqual(m['~albumartists'], 'Pink Floyd')
        self.assertEqual(m['~albumartists_sort'], 'Pink Floyd')
        self.assertEqual(m['~releasecomment'], 'stereo')
        self.assertEqual(m['~releaseannotation'], 'Original Vinyl release')
        self.assertEqual(m['~releaselanguage'], 'eng')
        self.assertEqual(m.getall('~releasecountries'), ['GB', 'NZ'])
        self.assertEqual(m['~release_series'], 'Why Pink Floyd?')
        self.assertEqual(m['~release_seriesid'], '7421b602-a413-4151-bcf4-d831debc3f27')
        self.assertEqual(m['~release_seriescomment'], 'Pink Floyed special editions')
        self.assertEqual(m['~release_seriesnumber'], '')
        self.assertEqual(a.genres, {
            'genre1': 6, 'genre2': 3,
            'tag1': 6, 'tag2': 3})
        for artist in a._album_artists:
            self.assertEqual(artist.genres, {
                'british': 2,
                'progressive rock': 10})

    def test_release_without_release_relationships(self):
        config.setting['release_ars'] = False
        m = Metadata()
        a = Album("1")
        release_to_metadata(self.json_doc, m, a)
        self.assertEqual(m['album'], 'The Dark Side of the Moon')
        self.assertEqual(m['albumartist'], 'Pink Floyd')
        self.assertEqual(m['albumartistsort'], 'Pink Floyd')
        self.assertEqual(m['asin'], 'b123')
        self.assertEqual(m['barcode'], '123')
        self.assertEqual(m['catalognumber'], 'SHVL 804')
        self.assertEqual(m['date'], '1973-03-24')
        self.assertEqual(m['label'], 'Harvest')
        self.assertEqual(m['musicbrainz_albumartistid'], '83d91898-7763-47d7-b03b-b92132375c47')
        self.assertEqual(m['musicbrainz_albumid'], 'b84ee12a-09ef-421b-82de-0441a926375b')
        self.assertEqual(m['producer'], '')
        self.assertEqual(m['releasecountry'], 'GB')
        self.assertEqual(m['releasestatus'], 'official')
        self.assertEqual(m['script'], 'Latn')
        self.assertEqual(m['~albumartists'], 'Pink Floyd')
        self.assertEqual(m['~albumartists_sort'], 'Pink Floyd')
        self.assertEqual(m['~releaselanguage'], 'eng')
        self.assertEqual(m.getall('~releasecountries'), ['GB', 'NZ'])
        self.assertEqual(a.genres, {
            'genre1': 6, 'genre2': 3,
            'tag1': 6, 'tag2': 3})
        for artist in a._album_artists:
            self.assertEqual(artist.genres, {
                'british': 2,
                'progressive rock': 10})

    def test_preferred_release_country(self):
        m = Metadata()
        a = Album("1")
        release_to_metadata(self.json_doc, m, a)
        self.assertEqual(m['releasecountry'], 'GB')
        config.setting['preferred_release_countries'] = ['NZ', 'GB']
        release_to_metadata(self.json_doc, m, a)
        self.assertEqual(m['releasecountry'], 'NZ')
        config.setting['preferred_release_countries'] = ['GB', 'NZ']
        release_to_metadata(self.json_doc, m, a)
        self.assertEqual(m['releasecountry'], 'GB')

    def test_media_formats_from_node(self):
        formats = media_formats_from_node(self.json_doc['media'])
        self.assertEqual(formats, '12" Vinyl')

    def test_release_group_rels(self):
        m = Metadata()
        release_group_to_metadata(self.json_doc['release-group'], m)
        self.assertEqual(m.getall('~releasegroup_series'), [
            "Absolute Radio's The 100 Collection",
            '1001 Albums You Must Hear Before You Die'
        ])
        self.assertEqual(m.getall('~releasegroup_seriesid'), [
            '4bf41050-6fa9-41a6-8398-15bdab4b0352',
            '4bc2a338-e1d8-4546-8a61-640da8aaf888'
        ])
        self.assertEqual(m.getall('~releasegroup_seriescomment'), [
            '2005 edition'
        ])
        self.assertEqual(m.getall('~releasegroup_seriesnumber'), ['15', '291'])

    def test_release_group_rels_double(self):
        m = Metadata()
        release_group_to_metadata(self.json_doc['release-group'], m)

        # load it twice and check for duplicates
        release_group_to_metadata(self.json_doc['release-group'], m)
        self.assertEqual(m.getall('~releasegroup_series'), [
            "Absolute Radio's The 100 Collection",
            '1001 Albums You Must Hear Before You Die'
        ])
        self.assertEqual(m.getall('~releasegroup_seriesid'), [
            '4bf41050-6fa9-41a6-8398-15bdab4b0352',
            '4bc2a338-e1d8-4546-8a61-640da8aaf888'
        ])
        self.assertEqual(m.getall('~releasegroup_seriescomment'), [
            '2005 edition'
        ])
        self.assertEqual(m.getall('~releasegroup_seriesnumber'), ['15', '291'])

    def test_release_group_rels_removed(self):
        m = Metadata()
        release_group_to_metadata(self.json_doc['release-group'], m)

        # remove one of the series from original metadata
        for i, rel in enumerate(self.json_doc['release-group']['relations']):
            if not rel['type'] == 'part of':
                continue
            if rel['series']['name'] == '1001 Albums You Must Hear Before You Die':
                del self.json_doc['release-group']['relations'][i]
                break
        release_group_to_metadata(self.json_doc['release-group'], m)
        self.assertEqual(m.getall('~releasegroup_series'), [
            "Absolute Radio's The 100 Collection",
        ])
        self.assertEqual(m.getall('~releasegroup_seriesid'), [
            '4bf41050-6fa9-41a6-8398-15bdab4b0352',
        ])
        self.assertEqual(m.getall('~releasegroup_seriescomment'), [])
        self.assertEqual(m.getall('~releasegroup_seriesnumber'), ['15'])


class NullReleaseTest(MBJSONTest):

    filename = 'release_null.json'

    def test_release(self):
        m = Metadata()
        a = Album("1")
        release_to_metadata(self.json_doc, m, a)
        self.assertEqual(m, {})

    def test_media_formats_from_node(self):
        formats = media_formats_from_node(self.json_doc['media'])
        self.assertEqual(formats, '(unknown)')


class RecordingTest(MBJSONTest):

    filename = 'recording.json'

    def test_recording(self):
        m = Metadata()
        t = Track('1')
        recording_to_metadata(self.json_doc, m, t)
        self.assertEqual(m['artist'], 'Ed Sheeran')
        self.assertEqual(m['artists'], 'Ed Sheeran')
        self.assertEqual(m['artistsort'], 'Sheeran, Ed')
        self.assertEqual(m['isrc'], 'GBAHS1400099')
        self.assertEqual(m['language'], 'eng')
        self.assertEqual(m['musicbrainz_artistid'], 'b8a7c51f-362c-4dcb-a259-bc6e0095f0a6')
        self.assertEqual(m['musicbrainz_recordingid'], 'cb2cc207-8125-445c-9ef9-6ea44eee959a')
        self.assertEqual(m['musicbrainz_workid'], 'dc469dc8-198e-42e5-b5a7-6be2f0a95ac0')
        self.assertEqual(m['performer:'], 'Ed Sheeran')
        self.assertEqual(m['performer:lead vocals'], 'Ed Sheeran')
        self.assertEqual(m['performer:guitar family'], 'Ed Sheeran')
        self.assertEqual(m['title'], 'Thinking Out Loud')
        self.assertEqual(m['work'], 'Thinking Out Loud')
        self.assertEqual(m['~workcomment'], 'Ed Sheeran song')
        self.assertEqual(m['writer'], 'Ed Sheeran; Amy Wadge')
        self.assertEqual(m['~writersort'], 'Sheeran, Ed; Wadge, Amy')
        self.assertEqual(m['~artists_sort'], 'Sheeran, Ed')
        self.assertEqual(m['~length'], '4:41')
        self.assertEqual(m['~recordingtitle'], 'Thinking Out Loud')
        self.assertEqual(m['~recording_firstreleasedate'], '2014-06-20')
        self.assertEqual(m['~video'], '')
        self.assertNotIn('originaldate', m)
        self.assertNotIn('originalyear', m)
        self.assertEqual(t.genres, {
            'blue-eyed soul': 1,
            'pop': 3})
        for artist in t._track_artists:
            self.assertEqual(artist.genres, {
                'dance-pop': 1,
                'guitarist': 0})

    def test_recording_instrument_credits(self):
        m = Metadata()
        t = Track('1')
        config.setting['standardize_instruments'] = False
        recording_to_metadata(self.json_doc, m, t)
        self.assertEqual(m['performer:vocals'], 'Ed Sheeran')
        self.assertEqual(m['performer:acoustic guitar'], 'Ed Sheeran')


class RecordingComposerCreditsTest(MBJSONTest):

    filename = 'recording_composer.json'

    def test_standardize_artists(self):
        m = Metadata()
        t = Track('1')
        config.setting['translate_artist_names'] = False
        config.setting['standardize_artists'] = True
        recording_to_metadata(self.json_doc, m, t)
        self.assertEqual(m['composer'], 'Пётр Ильич Чайковский')
        self.assertEqual(m['composersort'], 'Tchaikovsky, Pyotr Ilyich')

    def test_use_credited_as(self):
        m = Metadata()
        t = Track('1')
        config.setting['translate_artist_names'] = False
        config.setting['standardize_artists'] = False
        recording_to_metadata(self.json_doc, m, t)
        self.assertEqual(m['composer'], 'Tchaikovsky')
        self.assertEqual(m['composersort'], 'Tchaikovsky, Pyotr Ilyich')

    def test_translate(self):
        m = Metadata()
        t = Track('1')
        config.setting['translate_artist_names'] = True
        recording_to_metadata(self.json_doc, m, t)
        self.assertEqual(m['composer'], 'Pyotr Ilyich Tchaikovsky')
        self.assertEqual(m['composersort'], 'Tchaikovsky, Pyotr Ilyich')


class RecordingInstrumentalTest(MBJSONTest):

    filename = 'recording_instrumental.json'

    def test_recording(self):
        m = Metadata()
        t = Track('1')
        recording_to_metadata(self.json_doc, m, t)
        self.assertIn('instrumental', m.getall('~performance_attributes'))
        self.assertEqual(m['language'], 'zxx')
        self.assertNotIn('lyricist', m)


class MultiWorkRecordingTest(MBJSONTest):

    filename = 'recording_multiple_works.json'

    def test_recording(self):
        m = Metadata()
        t = Track('1')
        recording_to_metadata(self.json_doc, m, t)
        self.assertIn('instrumental', m.getall('~performance_attributes'))
        self.assertEqual(m['language'], 'jpn; eng; zxx')
        self.assertEqual(m['lyricist'], 'Satoru Kōsaki; Aki Hata; Minoru Shiraishi')
        self.assertEqual(m['~lyricistsort'], 'Kōsaki, Satoru; Hata, Aki; Shiraishi, Minoru')


class RecordingVideoTest(MBJSONTest):

    filename = 'recording_video.json'

    def test_recording(self):
        m = Metadata()
        t = Track('1')
        recording_to_metadata(self.json_doc, m, t)
        self.assertEqual(m['director'], 'Edward 209')
        self.assertEqual(m['producer'], 'Edward 209')
        self.assertEqual(m['~video'], '1')


class NullRecordingTest(MBJSONTest):

    filename = 'recording_null.json'

    def test_recording(self):
        m = Metadata()
        t = Track("1")
        recording_to_metadata(self.json_doc, m, t)
        self.assertEqual(m, {})


class RecordingCreditsTest(MBJSONTest):

    filename = 'recording_credits.json'

    def test_recording_solo_vocals(self):
        m = Metadata()
        t = Track("1")
        recording_to_metadata(self.json_doc, m, t)
        config.setting["standardize_artists"] = False
        self.assertNotIn('performer:solo', m)
        self.assertEqual(m['performer:solo vocals'], 'Frida')

    def test_recording_standardize_artist_credits(self):
        m = Metadata()
        t = Track("1")
        config.setting["standardize_artists"] = True
        recording_to_metadata(self.json_doc, m, t)
        self.assertNotIn('performer:solo', m)
        self.assertEqual(m['performer:solo vocals'], 'Anni-Frid Lyngstad')

    def test_recording_instrument_keep_case(self):
        m = Metadata()
        t = Track("1")
        recording_to_metadata(self.json_doc, m, t)
        self.assertEqual(m['performer:EWI'], 'Michael Brecker')


class TrackTest(MBJSONTest):

    filename = 'track.json'

    def test_track(self):
        t = Track("1")
        m = t.metadata
        track_to_metadata(self.json_doc, t)
        self.assertEqual(m['title'], 'Speak to Me')
        self.assertEqual(m['musicbrainz_recordingid'], 'bef3fddb-5aca-49f5-b2fd-d56a23268d63')
        self.assertEqual(m['musicbrainz_trackid'], 'd4156411-b884-368f-a4cb-7c0101a557a2')
        self.assertEqual(m['~length'], '1:08')
        self.assertEqual(m['tracknumber'], '1')
        self.assertEqual(m['~musicbrainz_tracknumber'], 'A1')
        self.assertEqual(m['~recordingcomment'], 'original stereo mix')
        self.assertEqual(m['~recordingtitle'], 'Speak to Me')


class PregapTrackTest(MBJSONTest):

    filename = 'track_pregap.json'

    def test_track(self):
        t = Track("1")
        m = t.metadata
        track_to_metadata(self.json_doc, t)
        self.assertEqual(m['title'], 'Lady')
        self.assertEqual(m['tracknumber'], '0')
        self.assertEqual(m['~musicbrainz_tracknumber'], '0')


class NullTrackTest(MBJSONTest):

    filename = 'track_null.json'

    def test_track(self):
        t = Track("1")
        m = t.metadata
        track_to_metadata(self.json_doc, t)
        self.assertEqual(m, {})


class MediaTest(MBJSONTest):

    filename = 'release_5medias.json'

    def test_media_formats_from_node_multi(self):
        formats = media_formats_from_node(self.json_doc['media'])
        self.assertEqual('2×CD + 2×DVD-Video + Blu-ray', formats)

    def test_medium_to_metadata_0(self):
        m = Metadata()
        medium_to_metadata(self.json_doc['media'][0], m)
        self.assertEqual(m['discnumber'], '1')
        self.assertEqual(m['media'], 'CD')
        self.assertEqual(m['totaltracks'], '5')
        self.assertEqual(m['discsubtitle'], 'The Original Album')

    def test_medium_to_metadata_4(self):
        m = Metadata()
        medium_to_metadata(self.json_doc['media'][4], m)
        self.assertEqual(m['discnumber'], '5')
        self.assertEqual(m['media'], 'Blu-ray')
        self.assertEqual(m['totaltracks'], '19')
        self.assertEqual(m['discsubtitle'], 'High Resolution Audio and Audio‐Visual Material')


class MediaPregapTest(MBJSONTest):

    filename = 'media_pregap.json'

    def test_track(self):
        m = Metadata()
        medium_to_metadata(self.json_doc, m)
        self.assertEqual(m['discnumber'], '1')
        self.assertEqual(m['media'], 'Enhanced CD')
        self.assertEqual(m['totaltracks'], '9')


class NullMediaTest(MBJSONTest):

    filename = 'media_null.json'

    def test_track(self):
        m = Metadata()
        medium_to_metadata(self.json_doc, m)
        self.assertEqual(m, {})


class NullArtistTest(MBJSONTest):

    filename = 'artist_null.json'

    def test_artist(self):
        m = Metadata()
        artist_to_metadata(self.json_doc, m)
        self.assertEqual(m, {})


class ArtistEndedTest(MBJSONTest):

    filename = 'artist_ended.json'

    def test_artist_ended(self):
        m = Metadata()
        artist_to_metadata(self.json_doc, m)
        self.assertEqual(m['area'], 'France')
        self.assertEqual(m['beginarea'], 'Paris')
        self.assertEqual(m['begindate'], '1928-04-02')
        self.assertEqual(m['endarea'], 'Paris')
        self.assertEqual(m['enddate'], '1991-03-02')
        self.assertEqual(m['gender'], 'Male')
        self.assertEqual(m['musicbrainz_artistid'], 'b21ef19b-c6aa-4775-90d3-3cc3e067ce6d')
        self.assertEqual(m['name'], 'Serge Gainsbourg')
        self.assertEqual(m['type'], 'Person')


class ArtistTranslationTest(MBJSONTest):

    filename = 'artist.json'

    def test_locale_specific_match_first(self):
        settings = {
            "standardize_tracks": False,
            "standardize_artists": False,
            "standardize_releases": False,
            "translate_artist_names": True,
            "translate_artist_names_script_exception": False,
            "standardize_instruments": True,
            "release_ars": True,
            "preferred_release_countries": [],
            "artist_locales": ['en_CA', 'en'],
        }
        self.set_config_values(settings)
        (artist_name, artist_sort_name) = _translate_artist_node(self.json_doc)
        self.assertEqual(artist_name, 'Ed Sheeran (en_CA)')

    def test_locale_specific_match_first_exc(self):
        settings = {
            "standardize_tracks": False,
            "standardize_artists": False,
            "standardize_releases": False,
            "translate_artist_names": True,
            "translate_artist_names_script_exception": True,
            "script_exceptions": [("LATIN", 0)],
            "standardize_instruments": True,
            "release_ars": True,
            "preferred_release_countries": [],
            "artist_locales": ['en_CA', 'en'],
        }
        self.set_config_values(settings)
        (artist_name, artist_sort_name) = _translate_artist_node(self.json_doc)
        self.assertEqual(artist_name, 'Ed Sheeran')

    def test_locale_specific_match_second(self):
        settings = {
            "standardize_tracks": False,
            "standardize_artists": False,
            "standardize_releases": False,
            "translate_artist_names": True,
            "translate_artist_names_script_exception": False,
            "standardize_instruments": True,
            "release_ars": True,
            "preferred_release_countries": [],
            "artist_locales": ['en_UK', 'en'],
        }
        self.set_config_values(settings)
        (artist_name, artist_sort_name) = _translate_artist_node(self.json_doc)
        self.assertEqual(artist_name, 'Ed Sheeran (en)')

    def test_artist_match_root_locale_fallback(self):
        settings = {
            "standardize_tracks": False,
            "standardize_artists": False,
            "standardize_releases": False,
            "translate_artist_names": True,
            "translate_artist_names_script_exception": False,
            "standardize_instruments": True,
            "release_ars": True,
            "preferred_release_countries": [],
            "artist_locales": ['en_UK'],
        }
        self.set_config_values(settings)
        (artist_name, artist_sort_name) = _translate_artist_node(self.json_doc)
        self.assertEqual(artist_name, 'Ed Sheeran (en)')

    def test_artist_no_match(self):
        settings = {
            "standardize_tracks": False,
            "standardize_artists": False,
            "standardize_releases": False,
            "translate_artist_names": True,
            "translate_artist_names_script_exception": False,
            "standardize_instruments": True,
            "release_ars": True,
            "preferred_release_countries": [],
            "artist_locales": ['de'],
        }
        self.set_config_values(settings)
        (artist_name, artist_sort_name) = _translate_artist_node(self.json_doc)
        self.assertEqual(artist_name, 'Ed Sheeran')


class ArtistTranslationArabicExceptionsTest(MBJSONTest):

    filename = 'artist_arabic.json'

    def test_locale_specific_match_first_exc1(self):
        settings = {
            "standardize_tracks": False,
            "standardize_artists": False,
            "standardize_releases": False,
            "translate_artist_names": True,
            "translate_artist_names_script_exception": True,
            "script_exceptions": [("LATIN", 0)],
            "standardize_instruments": True,
            "release_ars": True,
            "preferred_release_countries": [],
            "artist_locales": ['en_CA', 'en'],
        }
        self.set_config_values(settings)
        (artist_name, artist_sort_name) = _translate_artist_node(self.json_doc)
        self.assertEqual(artist_name, 'Mohamed Mounir')

    def test_locale_specific_match_first_exc2(self):
        settings = {
            "standardize_tracks": False,
            "standardize_artists": False,
            "standardize_releases": False,
            "translate_artist_names": True,
            "translate_artist_names_script_exception": True,
            "script_exceptions": [("ARABIC", 0)],
            "standardize_instruments": True,
            "release_ars": True,
            "preferred_release_countries": [],
            "artist_locales": ['en_CA', 'en'],
        }
        self.set_config_values(settings)
        (artist_name, artist_sort_name) = _translate_artist_node(self.json_doc)
        self.assertEqual(artist_name, 'محمد منير')


class TestAliasesLocales(PicardTestCase):

    def setUp(self):
        self.maxDiff = None

        self.aliases = [
            {
                "name": "Shearan",
                "sort-name": "Shearan",
                "primary": None,
                "locale": None,
                "type-id": ALIAS_TYPE_SEARCH_HINT_ID,
            },
            {
                "primary": True,
                "name": "Ed Sheeran (en)",
                "sort-name": "Sheeran, Ed",
                "type-id": ALIAS_TYPE_ARTIST_NAME_ID,
                "locale": "en",
            },
            {
                "primary": True,
                "name": "Ed Sheeran (en_CA)",
                "sort-name": "Sheeran, Ed",
                "type-id": ALIAS_TYPE_ARTIST_NAME_ID,
                "locale": "en_CA",
            },
        ]

    def test_1(self):
        expect_full = {'en': (0.8, ('Ed Sheeran (en)', 'Sheeran, Ed')), 'en_CA': (0.8, ('Ed Sheeran (en_CA)', 'Sheeran, Ed'))}
        expect_root = {'en': (0.8, ('Ed Sheeran (en)', 'Sheeran, Ed'))}

        full_locales, root_locales = _locales_from_aliases(self.aliases)
        self.assertDictEqual(expect_full, full_locales)
        self.assertDictEqual(expect_root, root_locales)

    def test_2(self):
        self.aliases[2]['type-id'] = ALIAS_TYPE_LEGAL_NAME_ID

        expect_full = {'en': (0.8, ('Ed Sheeran (en)', 'Sheeran, Ed')), 'en_CA': (0.65, ('Ed Sheeran (en_CA)', 'Sheeran, Ed'))}
        expect_root = {'en': (0.8, ('Ed Sheeran (en)', 'Sheeran, Ed'))}

        full_locales, root_locales = _locales_from_aliases(self.aliases)
        self.assertDictEqual(expect_full, full_locales)
        self.assertDictEqual(expect_root, root_locales)

    def test_3(self):
        self.aliases[0]['primary'] = True
        del self.aliases[0]['locale']

        expect_full = {'en': (0.8, ('Ed Sheeran (en)', 'Sheeran, Ed')), 'en_CA': (0.8, ('Ed Sheeran (en_CA)', 'Sheeran, Ed'))}
        expect_root = {'en': (0.8, ('Ed Sheeran (en)', 'Sheeran, Ed'))}

        full_locales, root_locales = _locales_from_aliases(self.aliases)
        self.assertDictEqual(expect_full, full_locales)
        self.assertDictEqual(expect_root, root_locales)

    def test_4(self):
        self.aliases[2]['type-id'] = ALIAS_TYPE_SEARCH_HINT_ID

        expect_full = {'en': (0.8, ('Ed Sheeran (en)', 'Sheeran, Ed')), 'en_CA': (0.4, ('Ed Sheeran (en_CA)', 'Sheeran, Ed'))}
        expect_root = {'en': (0.8, ('Ed Sheeran (en)', 'Sheeran, Ed'))}

        full_locales, root_locales = _locales_from_aliases(self.aliases)
        self.assertDictEqual(expect_full, full_locales)
        self.assertDictEqual(expect_root, root_locales)

    def test_5(self):
        self.aliases[1]['locale'] = 'en_US'
        self.aliases[1]['name'] = 'Ed Sheeran (en_US)'

        expect_full = {'en_US': (0.8, ('Ed Sheeran (en_US)', 'Sheeran, Ed')), 'en_CA': (0.8, ('Ed Sheeran (en_CA)', 'Sheeran, Ed'))}
        expect_root = {'en': (0.6, ('Ed Sheeran (en_US)', 'Sheeran, Ed'))}

        full_locales, root_locales = _locales_from_aliases(self.aliases)
        self.assertDictEqual(expect_full, full_locales)
        self.assertDictEqual(expect_root, root_locales)

    def test_6(self):
        self.aliases[2]['locale'] = 'en'
        self.aliases[2]['name'] = 'Ed Sheeran (en2)'
        self.aliases[2]['type-id'] = ALIAS_TYPE_ARTIST_NAME_ID
        self.aliases[1]['type-id'] = ALIAS_TYPE_LEGAL_NAME_ID
        self.aliases[1]['name'] = 'Ed Sheeran (en1)'

        expect_full = {'en': (0.8, ('Ed Sheeran (en2)', 'Sheeran, Ed'))}
        expect_root = {'en': (0.8, ('Ed Sheeran (en2)', 'Sheeran, Ed'))}

        full_locales, root_locales = _locales_from_aliases(self.aliases)
        self.assertDictEqual(expect_full, full_locales)
        self.assertDictEqual(expect_root, root_locales)


class ReleaseGroupTest(MBJSONTest):

    filename = 'release_group.json'

    def test_release_group(self):
        m = Metadata()
        r = ReleaseGroup("1")
        release_group_to_metadata(self.json_doc, m, r)
        self.assertEqual(m['musicbrainz_releasegroupid'], 'f5093c06-23e3-404f-aeaa-40f72885ee3a')
        self.assertEqual(m['~releasegroup_firstreleasedate'], '1973-03-24')
        self.assertEqual(m['originaldate'], '1973-03-24')
        self.assertEqual(m['originalyear'], '1973')
        self.assertEqual(m['releasetype'], 'album')
        self.assertEqual(m['~primaryreleasetype'], 'album')
        self.assertEqual(m['~releasegroup'], 'The Dark Side of the Moon')
        self.assertEqual(r.genres, {'test2': 3, 'test': 6})


class NullReleaseGroupTest(MBJSONTest):

    filename = 'release_group_null.json'

    def test_release_group(self):
        m = Metadata()
        r = ReleaseGroup("1")
        release_group_to_metadata(self.json_doc, m, r)
        self.assertEqual(m, {})


class CountriesFromNodeTest(MBJSONTest):

    filename = 'country.json'

    def test_countries_from_node(self):
        countries = countries_from_node(self.json_doc)
        self.assertEqual(['GB'], countries)

    def test_countries_from_node_no_event(self):
        del self.json_doc["release-events"]
        countries = countries_from_node(self.json_doc)
        self.assertEqual([], countries)

    def test_countries_from_node_no_area(self):
        del self.json_doc["release-events"][0]["area"]
        countries = countries_from_node(self.json_doc)
        self.assertEqual([], countries)


class CountriesFromNodeNullTest(MBJSONTest):

    filename = 'country_null.json'

    def test_countries_from_node(self):
        countries = countries_from_node(self.json_doc)
        self.assertEqual(countries, [])


class DatesCountriesFromNodeTest(MBJSONTest):

    filename = 'country.json'

    def test_dates_countries_from_node(self):
        dates, countries = release_dates_and_countries_from_node(self.json_doc)
        self.assertEqual(['GB'], countries)
        self.assertEqual(['1986-03'], dates)

    def test_dates_countries_from_node_no_event(self):
        del self.json_doc["release-events"]
        dates, countries = release_dates_and_countries_from_node(self.json_doc)
        self.assertEqual([], countries)
        self.assertEqual([], dates)


class DatesCountriesFromNodeNullTest(MBJSONTest):

    filename = 'country_null.json'

    def test_dates_countries_from_node(self):
        dates, countries = release_dates_and_countries_from_node(self.json_doc)
        self.assertEqual(countries, [])
        self.assertEqual([''], dates)


class LabelInfoTest(MBJSONTest):

    filename = 'label_info.json'

    def _label_info(self, n):
        return label_info_from_node(self.json_doc['releases'][n]['label-info'])

    def test_label_info_from_node_0(self):
        self.assertEqual((['naïve'], ['NJ628311']), self._label_info(0))

    def test_label_info_from_node_1(self):
        self.assertEqual((['naïve'], []), self._label_info(1))

    def test_label_info_from_node_2(self):
        self.assertEqual((['naïve'], []), self._label_info(2))

    def test_label_info_from_node_3(self):
        self.assertEqual(([], ["[None]"]), self._label_info(3))


class NullLabelInfoTest(MBJSONTest):

    filename = 'label_info_null.json'

    def test_label_info_from_node_0(self):
        label_info = label_info_from_node(self.json_doc['releases'][0]['label-info'])
        self.assertEqual(label_info, ([], []))


class GetScoreTest(PicardTestCase):
    def test_get_score(self):
        for score, expected in ((42, 0.42), ('100', 1.0), (0, 0.0), (None, 1.0), ('', 1.0)):
            self.assertEqual(expected, get_score({'score': score}))

    def test_get_score_no_score(self):
        self.assertEqual(1.0, get_score({}))


class ParseAttributeTest(PicardTestCase):

    def test_1(self):
        attrs, reltype, attr_credits = ('guest', 'keyboard'), 'instrument', {'keyboard': 'keyboards'}
        result = _parse_attributes(attrs, reltype, attr_credits)
        expected = 'guest keyboards'
        self.assertEqual(expected, result)

    def test_2(self):
        attrs, reltype, attr_credits = (), 'vocal', {}
        result = _parse_attributes(attrs, reltype, attr_credits)
        expected = 'vocals'
        self.assertEqual(expected, result)

    def test_3(self):
        attrs, reltype, attr_credits = ('guitar', 'keyboard'), 'instrument', {'keyboard': 'keyboards', 'guitar': 'weird guitar'}
        result = _parse_attributes(attrs, reltype, attr_credits)
        expected = 'weird guitar and keyboards'
        self.assertEqual(expected, result)


class RelationsToMetadataTargetTypeUrlTest(PicardTestCase):
    def test_invalid_asin_url(self):
        m = Metadata()
        relation = {
            'type': 'amazon asin',
            'url': {
                'resource': 'http://www.amazon.com/dp/020530902x',
            }
        }
        _relations_to_metadata_target_type_url(relation, m, None)
        self.assertEqual('', m['asin'])

    def test_has_asin_already(self):
        m = Metadata({'asin': 'ASIN'})
        relation = {
            'type': 'amazon asin',
            'url': {
                'resource': 'http://www.amazon.com/dp/020530902X',
            }
        }
        _relations_to_metadata_target_type_url(relation, m, None)
        self.assertEqual('ASIN', m['asin'])

    def test_valid_asin_url(self):
        m = Metadata()
        relation = {
            'type': 'amazon asin',
            'url': {
                'resource': 'http://www.amazon.com/dp/020530902X',
            }
        }
        _relations_to_metadata_target_type_url(relation, m, None)
        self.assertEqual('020530902X', m['asin'])

    def test_license_url(self):
        m = Metadata()
        relation = {
            'type': 'license',
            'url': {
                'resource': 'https://URL.LICENSE',
            }
        }
        _relations_to_metadata_target_type_url(relation, m, None)
        self.assertEqual('https://URL.LICENSE', m['license'])



# -*- coding: utf-8 -*-
#
# Picard, the next-generation MusicBrainz tagger
#
# Copyright (C) 2020 Laurent Monin
# Copyright (C) 2020, 2022 Philipp Wolfer
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.


from unittest.mock import (
    MagicMock,
    Mock,
)

from test.picardtestcase import PicardTestCase

from picard.acoustid.manager import (
    FINGERPRINT_MAX_ALLOWED_LENGTH_DIFF_MS,
    AcoustIDManager,
    Submission,
)
from picard.file import File
from picard.metadata import Metadata


def mock_succeed_submission(*args, **kwargs):
    # Run the callback
    args[1]({}, None, None)


def mock_fail_submission(*args, **kwargs):
    # Run the callback with error arguments
    args[1]({}, MagicMock(), True)


FINGERPRINT_SIZE = 4000


def dummy_file(i):
    file = File('foo%d.flac' % i)
    file.acoustid_fingerprint = 'Z' * FINGERPRINT_SIZE
    file.acoustid_length = 120
    file.metadata = Metadata(length=file.acoustid_length * 1000)
    return file


class AcoustIDManagerTest(PicardTestCase):
    def setUp(self):
        super().setUp()
        self.set_config_values({
            "clear_existing_tags": False,
            "compare_ignore_tags": []
        })
        self.mock_api_helper = MagicMock()
        self.mock_api_helper.submit_acoustid_fingerprints = Mock(wraps=mock_succeed_submission)
        self.acoustidmanager = AcoustIDManager(self.mock_api_helper)
        self.tagger.window = MagicMock()
        self.tagger.window.enable_submit = MagicMock()
        AcoustIDManager.MAX_PAYLOAD = FINGERPRINT_SIZE * 5
        AcoustIDManager.MAX_ATTEMPTS = 3

    def _add_unsubmitted_files(self, count):
        files = []
        for i in range(0, count):
            file = dummy_file(i)
            files.append(file)
            self.acoustidmanager.add(file, None)
            self.acoustidmanager.update(file, '00000000-0000-0000-0000-%012d' % i)
            self.assertFalse(self.acoustidmanager.is_submitted(file))
        return files

    def test_add_invalid(self):
        file = File('foo.flac')
        self.acoustidmanager.add(file, '00000000-0000-0000-0000-000000000001')
        self.tagger.window.enable_submit.assert_not_called()

    def test_add_and_update(self):
        file = dummy_file(0)
        self.acoustidmanager.add(file, '00000000-0000-0000-0000-000000000001')
        self.tagger.window.enable_submit.assert_called_with(False)
        self.acoustidmanager.update(file, '00000000-0000-0000-0000-000000000002')
        self.tagger.window.enable_submit.assert_called_with(True)
        self.acoustidmanager.update(file, '00000000-0000-0000-0000-000000000001')
        self.tagger.window.enable_submit.assert_called_with(False)

    def test_add_and_remove(self):
        file = dummy_file(0)
        self.acoustidmanager.add(file, '00000000-0000-0000-0000-000000000001')
        self.tagger.window.enable_submit.assert_called_with(False)
        self.acoustidmanager.update(file, '00000000-0000-0000-0000-000000000002')
        self.tagger.window.enable_submit.assert_called_with(True)
        self.acoustidmanager.remove(file)
        self.tagger.window.enable_submit.assert_called_with(False)

    def test_is_submitted(self):
        file = dummy_file(0)
        self.assertTrue(self.acoustidmanager.is_submitted(file))
        self.acoustidmanager.add(file, '00000000-0000-0000-0000-000000000001')
        self.assertTrue(self.acoustidmanager.is_submitted(file))
        self.acoustidmanager.update(file, '00000000-0000-0000-0000-000000000002')
        self.assertFalse(self.acoustidmanager.is_submitted(file))
        self.acoustidmanager.update(file, '')
        self.assertTrue(self.acoustidmanager.is_submitted(file))

    def test_submit_single_batch(self):
        f = self._add_unsubmitted_files(1)[0]
        self.acoustidmanager.submit()
        self.assertEqual(self.mock_api_helper.submit_acoustid_fingerprints.call_count, 1)
        self.assertEqual(
            f.acoustid_fingerprint,
            self.mock_api_helper.submit_acoustid_fingerprints.call_args[0][0][0].fingerprint
        )

    def test_submit_multi_batch(self):
        files = self._add_unsubmitted_files(int(AcoustIDManager.MAX_PAYLOAD / FINGERPRINT_SIZE) * 2)
        self.acoustidmanager.submit()
        self.assertEqual(self.mock_api_helper.submit_acoustid_fingerprints.call_count, 3)
        for f in files:
            self.assertTrue(self.acoustidmanager.is_submitted(f))

    def test_submit_multi_batch_failure(self):
        self.mock_api_helper.submit_acoustid_fingerprints = Mock(wraps=mock_fail_submission)
        files = self._add_unsubmitted_files(int(AcoustIDManager.MAX_PAYLOAD / FINGERPRINT_SIZE) * 2)
        self.acoustidmanager.submit()
        self.assertEqual(self.mock_api_helper.submit_acoustid_fingerprints.call_count, 8)
        for f in files:
            self.assertFalse(self.acoustidmanager.is_submitted(f))


class SubmissionTest(PicardTestCase):

    def test_init(self):
        fingerprint = 'abc'
        duration = 42
        recordingid = 'rec1'
        metadata = Metadata({
            'musicip_puid': 'puid1'
        })
        submission = Submission(fingerprint, duration, recordingid, metadata)
        self.assertEqual(fingerprint, submission.fingerprint)
        self.assertEqual(duration, submission.duration)
        self.assertEqual(recordingid, submission.recordingid)
        self.assertEqual(recordingid, submission.orig_recordingid)
        self.assertEqual(metadata, submission.metadata)
        self.assertEqual(metadata['musicip_puid'], submission.puid)
        self.assertEqual(0, submission.attempts)

    def test_bool(self):
        self.assertTrue(bool(Submission('foo', 1)))
        self.assertTrue(bool(Submission('foo', 0)))
        self.assertFalse(bool(Submission('foo', None)))
        self.assertFalse(bool(Submission(None, 1)))
        self.assertFalse(bool(Submission('', 1)))

    def test_valid_duration(self):
        duration_s = 342
        duration_ms = duration_s * 1000
        metadata = Metadata()
        submission = Submission('abc', duration_s, metadata=metadata)
        self.assertFalse(submission.valid_duration)
        metadata.length = duration_ms
        self.assertTrue(submission.valid_duration)
        metadata.length = duration_ms + FINGERPRINT_MAX_ALLOWED_LENGTH_DIFF_MS
        self.assertTrue(submission.valid_duration)
        metadata.length = duration_ms - FINGERPRINT_MAX_ALLOWED_LENGTH_DIFF_MS
        self.assertTrue(submission.valid_duration)
        metadata.length = duration_ms + 1 + FINGERPRINT_MAX_ALLOWED_LENGTH_DIFF_MS
        self.assertFalse(submission.valid_duration)
        metadata.length = duration_ms - 1 - FINGERPRINT_MAX_ALLOWED_LENGTH_DIFF_MS
        self.assertFalse(submission.valid_duration)

    def test_valid_duration_no_metadata(self):
        submission = Submission('abc', 342)
        self.assertTrue(submission.valid_duration)

    def test_init_no_metadata(self):
        submission = Submission('abc', 42)
        self.assertIsNone(submission.metadata)
        self.assertEqual('', submission.puid)

    def test_is_submitted_no_recording_id(self):
        submission = Submission('abc', 42)
        self.assertTrue(submission.is_submitted)
        submission.recordingid = 'rec1'
        self.assertFalse(submission.is_submitted)

    def test_is_submitted_move_recording_id(self):
        submission = Submission('abc', 42, recordingid='rec1')
        self.assertTrue(submission.is_submitted)
        submission.recordingid = 'rec2'
        self.assertFalse(submission.is_submitted)

    def test_args_with_mbid(self):
        submission = Submission('abc', 42, recordingid='rec1')
        expected = {
            'fingerprint': 'abc',
            'duration': '42',
            'mbid': 'rec1',
        }
        self.assertEqual(expected, submission.args)

    def test_args_with_mbid_with_puid(self):
        metadata = Metadata(
            musicip_puid='p1'
        )
        metadata.length = 42000
        submission = Submission('abc', 42, recordingid='rec1', metadata=metadata)
        expected = {
            'fingerprint': 'abc',
            'duration': '42',
            'mbid': 'rec1',
            'puid': 'p1'
        }
        self.assertEqual(expected, submission.args)

    def test_args_with_invalid_duration(self):
        metadata = Metadata({
            'title': 'The Track',
            'artist': 'The Artist',
            'album': 'The Album',
            'albumartist': 'The Album Artist',
            'tracknumber': '4',
            'discnumber': '2',
            'date': '2022-01-22',
        })
        metadata.length = 500000
        submission = Submission('abc', 42, recordingid='rec1', metadata=metadata)
        expected = {
            'fingerprint': 'abc',
            'duration': '42',
            'track': metadata['title'],
            'artist': metadata['artist'],
            'album': metadata['album'],
            'albumartist': metadata['albumartist'],
            'trackno': metadata['tracknumber'],
            'discno': metadata['discnumber'],
            'year': '2022',
        }
        self.assertEqual(expected, submission.args)

    def test_args_without_mbid(self):
        metadata = Metadata({
            'title': 'The Track',
            'artist': 'The Artist',
            'album': 'The Album',
            'albumartist': 'The Album Artist',
            'tracknumber': '4',
            'discnumber': '2',
            'date': '2022-01-22',
        })
        metadata.length = 42000
        submission = Submission('abc', 42, recordingid=None, metadata=metadata)
        expected = {
            'fingerprint': 'abc',
            'duration': '42',
            'track': metadata['title'],
            'artist': metadata['artist'],
            'album': metadata['album'],
            'albumartist': metadata['albumartist'],
            'trackno': metadata['tracknumber'],
            'discno': metadata['discnumber'],
            'year': '2022',
        }
        self.assertEqual(expected, submission.args)

    def test_args_year(self):
        metadata = Metadata({
            'year': '2022',
        })
        metadata.length = 500000
        submission = Submission('abc', 42, recordingid='rec1', metadata=metadata)
        args = submission.args
        self.assertEqual('2022', args['year'])

    def test_args_invalid_year(self):
        metadata = Metadata({
            'year': 'NaN',
        })
        metadata.length = 500000
        submission = Submission('abc', 42, recordingid='rec1', metadata=metadata)
        self.assertNotIn('year', submission.args)

    def test_len_mbid_puid(self):
        fingerprint = 'abc' * 30
        puid = 'p1'
        recordingid = 'rec1'
        duration = 42
        metadata = Metadata(
            musicip_puid=puid
        )
        metadata.length = 42000
        submission = Submission(fingerprint, 42, recordingid=recordingid, metadata=metadata)
        expected_min_length = len('&fingerprint=%s&duration=%s&mbid=%s&puid=%s' % (fingerprint, duration, recordingid, puid))
        self.assertGreater(len(submission), expected_min_length)

    def test_len_no_mbid(self):
        metadata = Metadata({
            'title': 'The Track',
            'artist': 'The Artist',
            'album': 'The Album',
            'albumartist': 'The Album Artist',
            'tracknumber': '4',
            'discnumber': '2',
            'date': '2022-01-22',
        })
        metadata.length = 500000
        submission = Submission('abc', 42, recordingid='rec1', metadata=metadata)
        expected_args = {
            'fingerprint': 'abc',
            'duration': '42',
            'track': metadata['title'],
            'artist': metadata['artist'],
            'album': metadata['album'],
            'albumartist': metadata['albumartist'],
            'trackno': metadata['tracknumber'],
            'discno': metadata['discnumber'],
            'year': '2022',
        }
        expected_min_length = len('&'.join(('='.join([k, v]) for k, v in expected_args.items())))
        self.assertGreater(len(submission), expected_min_length)
